parser grammar MathParser;

options {
  output = AST;
  tokenVocab = MathLexer;
}

tokens {
  COMBINE;
  DEFINE;
  DERIVATIVE;
  FUNCTION;
  POLYNOMIAL;
  TERM;
}

@header { package com.ociweb.math; }

@members { public boolean interactiveMode; }

assign
  : NAME ASSIGN value terminator -> ^(ASSIGN NAME value);

coefficient: NUMBER;

combine
  : fn1=NAME ASSIGN fn2=NAME op=SIGN fn3=NAME terminator
    -> ^(COMBINE $fn1 $op $fn2 $fn3);

define
  : fn=NAME LEFT_PAREN fv=NAME RIGHT_PAREN ASSIGN
    polynomial[$fn.text, $fv.text] terminator
    -> ^(DEFINE $fn $fv polynomial);
  
derivative
  : NAME APOSTROPHE LEFT_PAREN RIGHT_PAREN -> ^(DERIVATIVE NAME);

exponent: CARET NUMBER -> NUMBER;

functionEval
  : fn=NAME LEFT_PAREN (v=NUMBER | v=NAME) RIGHT_PAREN -> ^(FUNCTION $fn $v);
  
help: HELP terminator -> HELP;
  
interactiveStatement: help | list;
  
list
  : LIST listOption terminator -> ^(LIST listOption);

listOption: FUNCTIONS | VARIABLES;

// fnt = function name text; fvt = function variable text
polynomial[String fnt, String fvt]
  : term[$fnt, $fvt] (SIGN term[$fnt, $fvt])*
    -> ^(POLYNOMIAL term (SIGN term)*);
  
print
  : PRINT printTarget* terminator -> ^(PRINT printTarget*);
  
printTarget
  : NUMBER -> NUMBER
  | sl=STRING_LITERAL -> $sl
  | NAME -> NAME
  // The next alternative is a function reference
  // to print a string representation.
  // It isn't needed if a disambiguating predicate is used in
  // the tree grammar.  See comments in tree grammar print target.
  | NAME LEFT_PAREN RIGHT_PAREN -> ^(FUNCTION NAME)
  | functionEval
  | derivative
  ;

// This is the "start rule".
script: statement*; // EOF!;

statement
  : assign
  | combine
  | define
  // The next line demonstrates using a gated semantic predicate.
  | /* { interactiveMode }?=> */ interactiveStatement
  | print
  ;

// fnt = function name text; fvt = function variable text
term[String fnt, String fvt]
  // tv = term variable
  : c=coefficient? (tv=NAME e=exponent?)?
    // What follows is a validating semantic predicate.
    // If it evaluates to false, a FailedPredicateException will be thrown.
    { tv == null ? true : ($tv.text).equals($fvt) }?
    -> ^(TERM $c? $tv? $e?)
  ;
  // TODO: Do this in override of getErrorMessage instead?
  catch [FailedPredicateException fpe] {
    String tvt = $tv.text;
    String msg = "In function \"" + fnt +
      "\" the term variable \"" + tvt +
      "\" doesn't match function variable \"" + fvt + "\".";
    throw new RuntimeException(msg);
  }

// EOF cannot be used in lexer rules, so we made this a parser rule.
// EOF is needed here for interactive mode where each line entered ends in EOF.
terminator: NEWLINE | EOF;

value: NUMBER | NAME | functionEval;
